; -----------------------------------------------------------------------------
; MODULE "MICRO BUG EXTRA"
;
; AUTHOR: EINAR GATTONI SAUKAS
; -----------------------------------------------------------------------------

; -----------------------------------------------------------------------------
; MICRO BUG routines

MB_SCRUP            EQU     $7860       ; scroll up
MB_SCRDN            EQU     $7879       ; scroll down
MB_CONT             EQU     $78AA       ; wait keyboard release
MB_PRINT            EQU     $78C1       ; print message
MB_PRINT_INLINE     EQU     $78C7       ; print message without scroll
MB_KEY              EQU     $78D3       ; read key
MB_EXIT             EQU     $7A4A       ; exit to main menu
MB_CURSOR           EQU     $7A52       ; read string
MB_VALDE            EQU     $7A9E       ; parse first numeric parameter
MB_VALDE_NEXT       EQU     $7AA7       ; parse next numeric parameter
MB_ENDER            EQU     $7B02       ; print memory address
MB_ENDER_NOFLAG     EQU     $7B0C       ; print memory address without flags
MB_ENDER_DEC        EQU     $7B1B       ; print memory address in decimal
MB_AVNM_SCRUP       EQU     $7B52       ; print command M row with scroll up
MB_AVNM_NOSCR       EQU     $7B55       ; print command M row without scroll up
MB_COMMAND_M        EQU     $7B7B       ; command M
MB_ORG              EQU     $7BF7       ; read command parameters
MB_ORG_PARSE        EQU     $7C00       ; parse command parameters
MB_LINHA_ILEGAL     EQU     $7C27       ; print 'LINHA ILEGAL' and exit
MB_SLOW_EXIT        EQU     $7C36       ; reset mode SLOW and exit
MB_BLOCO_ILEGAL     EQU     $7CAE       ; print 'BLOCO ILEGAL'
MB_COMMAND_O_COPY   EQU     $7CC2       ; copy memory area (LDIR or LDDR as needed)
MB_COMMAND_D        EQU     $7D27       ; command D
MB_PROX             EQU     $7E4B       ; increment address, exit if finished
MB_LOAD_BYTE        EQU     $7EA1       ; load single byte from tape
MB_IVG              EQU     $7EF5       ; input tape parameters, adjust speed
MB_SOMA             EQU     $7FC9       ; calculate block sum


; -----------------------------------------------------------------------------
; MICRO BUG variables

MB_MFLAG            EQU     $7926       ; control flags MFLAG
MB_PARAM1           EQU     $7931       ; 1st input parameter (word)
MB_PARAM2           EQU     $7933       ; 2nd input parameter (word)
MB_PARAM3           EQU     $7935       ; 3rd input parameter (byte)


; -----------------------------------------------------------------------------
; unused area

org $6E00
                    DB $00


; -----------------------------------------------------------------------------
; command S: display memory sums (enhanced variant of command P)
;            "S [address][,step]"
; -----------------------------------------------------------------------------

org $7500
S_SUM:
                    XOR A               ; print space
                    RST $10
                    CALL MB_CURSOR      ; read string
                    OR A                ; 1st parameter?
                    JR Z,S_NOPARAM
                    CALL MB_ORG_PARSE   ; parse command parameters
                    LD HL,(MB_PARAM1)
                    LD (VAL_S_ADDR),HL  ; store new address
                    LD A,D              ; 2nd parameter?
                    OR E
                    JR Z,S_NOPARAM
                    LD (VAL_S_STEP),DE  ; store new step
S_NOPARAM:
                    CALL MB_SCRUP       ; scroll up
S_FORWARD:
                    CALL AVNS           ; move forward
S_LOOP:
                    CALL MB_KEY         ; read key
                    OR A                ; return if BREAK
                    RET Z
                    CP $30              ; forward if "K"
                    JR Z,S_FORWARD
                    CP $2F              ; backward if "J"
                    CALL Z,RTRS         ; move backward
                    JR S_LOOP


; -----------------------------------------------------------------------------
; routine RTRS (move backward for command S)

RTRS:
                    LD HL,(VAL_S_ADDR)  ; read start address
                    LD DE,(VAL_S_STEP)  ; read step
                    PUSH DE
                    AND A               ; decrement start address
                    SBC HL,DE
                    LD (VAL_S_ADDR),HL  ; store new start address
                    LD B,$16            ; calculate address at top of screen
RTRS_LOOP:
                    AND A
                    SBC HL,DE
                    DJNZ RTRS_LOOP
                    CALL MB_SCRDN       ; scroll down
                    JR AVNS_ROW         ; print memory address and sum


; -----------------------------------------------------------------------------
; routine AVNS (move forward for command S)

AVNS:
                    LD HL,(VAL_S_ADDR)  ; read start address
                    LD DE,(VAL_S_STEP)  ; read step
                    PUSH DE
                    PUSH HL
                    ADD HL,DE           ; increment start address
                    LD (VAL_S_ADDR),HL  ; store new start address
                    POP HL
                    CALL MB_SCRUP       ; scroll up
AVNS_ROW:
                    CALL MB_ENDER       ; print memory address
                    POP DE
AVNS_SUM:
                    PUSH HL
                    ADD HL,DE           ; calculate end address
                    LD (MB_PARAM2),HL   ; prepare end address to calculate block sum
                    CALL MB_PRINT_INLINE ; print message without scroll
                    DB $16,$12,$80      ; '-> '
                    POP HL              ; prepare start address to calculate block sum
                    LD BC,$0000
                    CALL MB_SOMA        ; calculate block sum
                    LD HL,$7B16         ; patch routine $ENDER to stop printing spaces after hex value
                    LD (HL),$C9
                    PUSH HL
                    LD HL,$7B39         ; patch routine $ENDER to stop printing spaces after dec value
                    LD (HL),$18
                    PUSH HL
                    LD H,B              ; prepare to print block sum instead of address
                    LD L,C
                    LD A,(MB_MFLAG)     ; check control flags (MFLAG)
                    BIT 5,A             ; flag SHIFT+A ?
                    CALL MB_ENDER_NOFLAG ; print memory address without flags
                    POP HL              ; fix routine $ENDER
                    LD (HL),$2A
                    POP HL              ; fix routine $ENDER
                    LD (HL),$AF
                    RET


; -----------------------------------------------------------------------------
; command U: display memory blocks with sums (enhanced variant of command D)
;            "U [address]"
; -----------------------------------------------------------------------------

U_VIEW:
                    LD HL,UNPATCH       ; patch main MICRO BUG routine to fix itself afterwards
                    LD ($7A09),HL
                    XOR A               ; patch routine $AVND to print without spaces
                    LD ($7D0C),A
                    LD ($7CFD),A
                    LD HL,ROWU          ; patch routine $AVND to call routine ROWU
                    LD ($7CED),HL
                    JP MB_COMMAND_D     ; execute command D


; -----------------------------------------------------------------------------
; routine ROWU (print row for command U)

ROWU:
                    POP DE              ; discard return address
                    CALL MB_ENDER       ; print block address
                    PUSH HL             ; save block address
                    CALL $7CEF          ; execute remaining part of routine $AVND
                    POP HL              ; restore block address
                    XOR A
                    RST $10             ; print space
                    LD DE,8
                    JR AVNS_SUM         ; print block sum


; -----------------------------------------------------------------------------
; routine UNPATCH (fix all patches from commands U and W)

UNPATCH:
                    LD HL,MB_ENDER
                    LD ($7B56),HL       ; fix routine $AVNM
                    LD ($7CED),HL       ; fix routine $AVND
                    LD A,$D7
                    LD ($7CFD),A        ; fix $AVND to print space between blocks
                    LD ($7D0C),A        ; fix $AVND to print space between blocks
                    LD HL,MB_CONT
                    LD ($7A09),HL       ; fix main routine
                    JP (HL)


; -----------------------------------------------------------------------------
; command W: edit memory with CHR$ (enhanced variant of command M)
;            "W [address]"
; -----------------------------------------------------------------------------

W_EDIT:
                    LD HL,UNPATCH       ; patch main MICRO BUG routine to fix itself afterwards
                    LD ($7A09),HL
                    LD HL,ENDERW
                    LD ($7B56),HL       ; patch routine $AVNM to call routine ENDERW
                    JP MB_COMMAND_M     ; execute command M


; -----------------------------------------------------------------------------
; routine ENDERW (print address and CHR$ for command W)

ENDERW:
                    LD E,(HL)
                    PUSH HL
                    LD HL,($400E)       ; save PRINT AT position
                    LD B,$10
                    XOR A
ENDERW_LOOP:
                    RST $10             ; print 16 spaces
                    DJNZ ENDERW_LOOP
                    LD C,B
                    LD A,E
                    CALL $0B75          ; print CHR$ A
                    LD ($400E),HL       ; restore PRINT AT position
                    POP HL
                    JP MB_ENDER         ; print memory address and return


; -----------------------------------------------------------------------------
; command V: verify memory area saved on tape
;            "V start,end"
; -----------------------------------------------------------------------------
V_VERIFY:
                    CALL MB_IVG         ; input tape parameters, adjust speed
                    CALL VERIFY         ; verify
                    JP MB_SLOW_EXIT     ; reset mode SLOW and exit


; -----------------------------------------------------------------------------
; routine VERIFY (verify from tape)

VERIFY:
                    CALL $0F23          ; FAST mode
                    LD HL,(MB_PARAM1)   ; start address
VERIFY_LOOP:
                    LD D,B
                    CALL MB_LOAD_BYTE   ; load next byte
                    LD A,C
                    CP (HL)             ; matches next byte in memory?
                    JR Z,VERIFY_PROX
                    CALL MB_PRINT       ; print message
                    DB $2A,$37,$37,$B4  ; 'ERRO'
                    RET                 ; exit
VERIFY_PROX:
                    INC HL              ; next address
                    EX DE,HL
                    LD HL,(MB_PARAM2)   ; end address
                    SCF
                    SBC HL,DE
                    EX DE,HL
                    JR NC,VERIFY_CONT   ; continue if not reached end address
                    CALL MB_PRINT       ; print message
                    DB $34,$B0          ; 'OK'
                    RET                 ; exit
VERIFY_CONT:
                    JR VERIFY_LOOP      ; extra jump to match LOAD timing exactly


; -----------------------------------------------------------------------------
; command X: compare memory addresses to find next difference
;            "X [address1,address2]"
; -----------------------------------------------------------------------------

X_DIFF:
                    XOR A               ; print space
                    RST $10
                    CALL MB_CURSOR      ; read string
                    LD HL,(VAL_X_ADDR1) ; continue last execution by default
                    INC HL
                    LD DE,(VAL_X_ADDR2)
                    INC DE
                    OR A                ; 1st parameter?
                    JR Z,X_NOPARAM      ; use old parameters
                    CALL MB_ORG_PARSE   ; parse command parameters
                    LD HL,(MB_PARAM1)
                    ;AND A              ; (unnecessary)
                    SBC HL,DE
                    RET Z               ; exit if addresses are identical
                    ADD HL,DE
X_NOPARAM:
                    PUSH DE
                    CALL MB_SCRUP       ; scroll up
                    POP DE
X_LOOP:
                    LD (VAL_X_ADDR1),HL ; store current position
                    LD (VAL_X_ADDR2),DE
                    LD A,(DE)
                    CP (HL)             ; match content?
                    JR Z,X_NEXT
                    PUSH HL             ; preserve both addresses
                    PUSH DE
                    CALL MB_AVNM_SCRUP  ; print 1st address and byte content
                    LD BC,$150D
                    CALL $08F5          ; AT 21,13
                    LD A,$16            ; print '-  '
                    RST $10
                    XOR A
                    RST $10
                    RST $10
                    POP HL
                    PUSH HL
                    LD A,$19            ; patch routine $ENDER for different column
                    LD ($7B3D),A
                    CALL MB_AVNM_NOSCR  ; print 2nd address and byte content
                    LD A,$09            ; patch routine $ENDER for different column
                    LD ($7B3D),A
                    CALL MB_KEY         ; read key
                    POP DE
                    POP HL
                    OR A
                    RET Z               ; exit if BREAK
X_NEXT:
                    INC HL
                    INC DE
                    JR X_LOOP           ; find next mismatch


; -----------------------------------------------------------------------------
; command Y: find pattern in memory area
;            "Y [start][,end][,pattern]"
; -----------------------------------------------------------------------------

Y_FIND:
                    XOR A               ; print space
                    RST $10
                    CALL MB_CURSOR      ; read string
                    OR A                ; 1st parameter?
                    JR Z,Y_NOPARAM      ; if no parameters, continue last execution
                    CALL PATTERN        ; parse input string with pattern
                    LD (VAL_Y_PATTERN_SIZE),HL ; store new pattern size
                    EX DE,HL
                    JR C,Y_FIXED_PATTERN ; if pattern in fixed memory area, don't copy it
                    LD B,D
                    LD C,E
                    LD DE,VAL_Y_PATTER_BUFFER
                    PUSH DE
                    LDIR                ; otherwise copy pattern to internal buffer
                    POP HL
Y_FIXED_PATTERN:
                    LD (VAL_Y_PATTERN_ADDR),HL ; store new pattern address
                    LD HL,(MB_PARAM1)   ; new block address
                    LD (VAL_Y_BLOCK_ADDR),HL ; save new block address
                    EX DE,HL
                    LD HL,(MB_PARAM2)   ; new block size
                    LD A,H
                    OR L                ; provided block size?
                    JR NZ,Y_BLOCK_SIZE
                    SBC HL,DE           ; calculate block size until end of memory
                    JR NZ,Y_BLOCK_SIZE
                    DEC HL
Y_BLOCK_SIZE:
                    LD (VAL_Y_BLOCK_SIZE),HL ; save new block size
Y_NOPARAM:
                    CALL MB_SCRUP       ; scroll up
Y_MATCH_FIRST:
                    LD BC,(VAL_Y_BLOCK_SIZE)
                    LD A,B
                    OR C
                    RET Z               ; exit if search finished
                    LD HL,(VAL_Y_BLOCK_ADDR)
                    LD DE,(VAL_Y_PATTERN_ADDR)
                    LD A,(DE)
                    CPIR                ; search for first byte from pattern
                    RET NZ              ; exit if nothing found
                    LD (VAL_Y_BLOCK_ADDR),HL ; store next search address
                    DEC HL              ; restore match address
                    LD (VAL_Y_BLOCK_SIZE),BC ; store remaining block size
                    LD BC,(VAL_Y_PATTERN_SIZE) ; restore pattern size
Y_MATCH_REST:
                    LD A,(DE)
                    CPI                 ; compare each byte from pattern
                    JR NZ,Y_MATCH_FIRST ; if mismatch then restart
                    INC DE
                    JP PE,Y_MATCH_REST  ; loop until all bytes are compared
                    CALL MB_SCRUP       ; scroll up
                    LD A,$14            ; print '='
                    RST $10
                    LD HL,$7B3D         ; patch routine $ENDER for different column
                    INC (HL)
                    PUSH HL
                    LD HL,(VAL_Y_BLOCK_ADDR)
                    DEC HL              ; restore match address
                    CALL $7CEC          ; print match address
                    POP HL
                    DEC (HL)            ; fix routine $ENDER
                    CALL MB_KEY         ; read key
                    OR A
                    JR NZ,Y_MATCH_FIRST ; repeat until BREAK
                    RET


; -----------------------------------------------------------------------------
; command Z: fill memory area with pattern (enhanced variant of command F)
;            "Z start[,end][,pattern]"
; -----------------------------------------------------------------------------
Z_FILL:
                    XOR A               ; print space
                    RST $10
                    CALL MB_CURSOR      ; read string
                    OR A                ; 1st parameter?
                    RET Z
                    CALL PATTERN        ; parse input string with pattern
                    LD B,H
                    LD C,L              ; pattern size
                    LD HL,(MB_PARAM2)   ; memory area size
                    LD A,H
                    OR L                ; provided memory area size?
                    JR Z,Z_COPY_ALL     ; if not, copy entire pattern to start address
                    PUSH HL             ; save memory area size
                    SCF
                    SBC HL,BC           ; memory area size > pattern size?
                    JR C,Z_COPY_PART    ; if not, copy part of pattern to fill area
                    INC HL              ; calculate difference "memory area size - pattern size"
                    PUSH HL             ; save difference "memory area size - pattern size"
                    LD HL,(MB_PARAM1)   ; memory area start address
                    EX DE,HL            ; prepare to copy pattern to memory area start address
                    PUSH BC             ; save pattern size
                    PUSH DE             ; save memory area start address
                    CALL MB_COMMAND_O_COPY ; copy pattern to beginning of memory area
                    POP DE              ; restore memory area start address
                    POP HL              ; restore pattern size
                    ADD HL,DE           ; calculate start address of remaining memory area
                    EX DE,HL            ; prepare to copy from beginning to remaining of memory area
                    POP BC              ; restore difference "memory area size - pattern size"
                    LDIR                ; fill remaining memory area copying from itself
                    JR T_MATCHES        ; print memory area size (in decimal) and exit
Z_COPY_PART:
                    POP BC              ; restore memory area size (which is <= pattern size)
Z_COPY_ALL:
                    push bc             ; save memory area size
                    LD HL,(MB_PARAM1)   ; memory area start address
                    EX DE,HL            ; prepare to copy from pattern to memory area
                    CALL MB_COMMAND_O_COPY ; copy (part of) pattern to memory area
                    JR T_MATCHES        ; print memory area size (in decimal) and exit


; -----------------------------------------------------------------------------
; command T: replace byte code within memory area
;            "T start,end,cc,dd"
; -----------------------------------------------------------------------------

T_REPLACE:
                    CALL MB_ORG         ; read command parameters
                    PUSH DE             ; save end address
                    CALL MB_VALDE_NEXT  ; parse next numeric parameter
                    LD A,E
                    PUSH AF             ; save search code
                    INC HL
                    CALL MB_VALDE_NEXT  ; parse next numeric parameter
                    LD A,E
                    LD (REPLACE_SET+1),A ; store replace code
                    POP AF              ; restore search code
                    POP DE              ; restore end address
                    LD HL,(MB_PARAM1)   ; start address
                    ;AND A              ; (unnecessary)
                    SBC HL,DE           ; error if start address >= end address
                    JR NC,BLOCO_ILEGAL  ; print block error and exit
                    ADD HL,DE
                    LD BC,$0000         ; match count
                    CALL REPLACE_START  ; replace matches
                    PUSH BC             ; save match count
T_MATCHES:
                    POP HL              ; restore match count
                    CALL MB_SCRUP       ; scroll up
                    CALL MB_SCRUP       ; scroll up
                    JP MB_ENDER_DEC     ; print match count (in decimal) and exit


; -----------------------------------------------------------------------------
; routine REPLACE (replace matches)

REPLACE_LOOP:
                    CALL MB_PROX        ; increment address, exit if finished
REPLACE_START:
                    CP (HL)             ; match?
                    JR NZ,REPLACE_LOOP  ; continue otherwise
REPLACE_SET:
                    LD (HL),$00         ; replace
                    INC BC              ; increment match count
                    JR REPLACE_LOOP     ; continue


; -----------------------------------------------------------------------------
; routine PATTERN (parse input string with pattern)

PATTERN:
                    CALL MB_ORG_PARSE   ; parse command parameters
                    LD A,D              ; end address?
                    OR E
                    JR Z,PATTERN_NOSIZE ; leave zero if unknown end address
                    PUSH HL             ; preserve parsing address
                    LD HL,(MB_PARAM1)   ; start address
                    EX DE,HL
                    SCF
                    SBC HL,DE           ; calculate block size - 1
                    JR C,BLOCO_ILEGAL   ; error if start address >= end address
                    INC HL              ; calculate actual block size
                    LD (MB_PARAM2),HL   ; replace 2nd parameter with block size
                    POP HL              ; restore parsing address
PATTERN_NOSIZE:
                    LD A,(HL)
                    OR A                ; 1st pattern parameter?
                    JR Z,INPUT_STRING
                    CALL MB_VALDE_NEXT  ; parse next numeric parameter
                    OR A
                    JR NZ,PATTERN_BLOCK ; 2nd pattern parameter?
PATTERN_BYTE:
                    LD A,E              ; pattern is a single byte
                    LD DE,MB_PARAM3
                    ld (DE),A           ; store single byte
                    ld HL,1             ; pattern size = 1
                    RET
PATTERN_BLOCK:
                    PUSH DE             ; save pattern start address
                    INC HL
                    CALL MB_VALDE_NEXT  ; parse next numeric parameter
                    EX DE,HL            ; pattern end address
                    POP DE              ; restore pattern start address
                    SCF
                    SBC HL,DE           ; calculate pattern size - 1
                    INC HL              ; calculate actual pattern size
                    CCF                 ; use carry flag to indicate pattern block
                    RET C               ; error if pattern start >= pattern end


; -----------------------------------------------------------------------------
; routine BLOCO_ILEGAL (print block error and exit)

BLOCO_ILEGAL:
                    CALL MB_BLOCO_ILEGAL ; print 'BLOCO ILEGAL'
                    JP MB_EXIT           ; exit


; -----------------------------------------------------------------------------
; routine INPUT_STRING (read input string)

INPUT_STRING:
                    CALL MB_SCRUP       ; scroll up
                    LD A,$0B
                    RST $10             ; print '"'
                    LD HL,($400E)       ; PRINT AT position
                    PUSH HL             ; save pattern start address
                    CALL MB_CURSOR      ; read string
                    POP DE              ; restore pattern start address
                    CP $76
                    JP Z,MB_LINHA_ILEGAL ; if invalid input string then print 'LINHA ILEGAL' and exit
                    LD A,$0B
                    RST $10             ; print '"'
                    ;AND A              ; (unnecessary)
                    SBC HL,DE           ; calculate input string size - 1
                    JP C,MB_LINHA_ILEGAL ; if empty input string then print 'LINHA ILEGAL' and exit
                    INC HL              ; calculate actual input string size
                    RET


; -----------------------------------------------------------------------------
; data area

VAL_S_ADDR:
                    DW $0000            ; S parameter "start"
VAL_S_STEP:
                    DW $0008            ; S parameter "step"
VAL_X_ADDR1:
                    DW $1DFF            ; X parameter "address1"
VAL_X_ADDR2:
                    DW $1E07            ; X parameter "address2"
VAL_Y_BLOCK_ADDR:
                    DW $0000            ; Y memory area start address
VAL_Y_BLOCK_SIZE:
                    DW $0000            ; Y memory area size
VAL_Y_PATTERN_ADDR:
                    DW $0000            ; Y pattern start address
VAL_Y_PATTERN_SIZE:
                    DW $0000            ; Y pattern size
VAL_Y_PATTER_BUFFER:
                    DS $1E              ; Y pattern string buffer (30 bytes)

; -----------------------------------------------------------------------------
; TCM: module commands table

org $7800
TCM:
                    DW S_SUM            ; S [address][,step]
                    DW T_REPLACE        ; T start,end,cc,dd
                    DW U_VIEW           ; U [address]
                    DW V_VERIFY         ; V start,end
                    DW W_EDIT           ; W [address]
                    DW X_DIFF           ; X [address1,address2]
                    DW Y_FIND           ; Y [start][,end][,pattern]
                    DW Z_FILL           ; Z start[,end][,pattern]

; -----------------------------------------------------------------------------
; routine $MORG (organize implementation of operating module)

MORG:
                    LD BC,$1200
                    CALL $08F5          ; AT 18,0
                    CALL MB_PRINT_INLINE ; print message without scroll
                    DB $32,$27,$00,$2A  ; 'MB EXTRA (EINAR)'
                    DB $3D,$39,$37,$26
                    DB $00,$10,$2A,$2E
                    DB $33,$26,$37,$91
                    LD HL,TCM           ; copy TCM to main commands table
                    LD DE,$7964
                    LD BC,$10
                    LDIR
                    JP $79E3            ; back to command loop

org $783A
                    DW $6350            ; module sum from "P 6E00,783A"

; -----------------------------------------------------------------------------
